////////////////////////////////////////////////////////////////////////
// Class:       OnlinePosition
// Module Type: producer
// File:        OnlinePosition_module.cc
//
// Generated at Mon Aug 19 10:52:46 2013 by Alessandro Razeto & Nicola Rossi using artmod
// from cetpkgsupport v1_02_00.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Utilities/InputTag.h"
#include "fhiclcpp/ParameterSet.h"
#include "darksidecore/ArtServices/DBInterface.hh"

#include "darkmon/ArtModules/OnlineEvent.hh"

#include <TF2.h>
#include <TMinuit.h>
#include <TGraph.h>
#include <TH1.h>
#include <TFile.h>

#include <numeric>
#include <vector>
#include <functional>
#include <algorithm>
#include <iostream>
#include <cmath>
#include <fstream>
#include <string>

#define SA_RINGS

namespace ds50 {
  class OnlinePosition : public art::EDProducer {
    public:
      explicit OnlinePosition(fhicl::ParameterSet const & p);
      virtual ~OnlinePosition();

      void produce(art::Event & e) override;
      void beginRun(art::Run &) override;

      double operator() (double *x, double *) const { return lkl(x[0], x[1]); }
      float lkl (float x, float y) const;

    private:
      float drift_;
      bool do_lkl_;
      std::string sa_calib_;

      static constexpr float PMT_R = 90/115.;
      static constexpr float TPC_R = 495/115.;

      struct charge_deposit {
	float x, y, z, r, c, nc;
      };

      charge_deposit cds_[40];
#ifdef SA_RINGS
      std::unique_ptr<TGraph> solid_angles_[4], topology_corrections_[4];
#else
      std::unique_ptr<TGraph> solid_angles_[40], topology_corrections_[4];
#endif

      std::unique_ptr<TGraph> inverse_cumulative (TH1* h);
      std::unique_ptr<TGraph> cumulative (TH1* h);
      float avoid_tgraph_insane_extrapolation (float x, const TGraph& g) const;
  };
}

ds50::OnlinePosition::OnlinePosition(fhicl::ParameterSet const & ps): drift_(ps.get<float> ("drift", 1)), do_lkl_(ps.get<bool> ("do_lkl", true)), sa_calib_(ps.get<std::string> ("sa_calib", "sa_calib.root")) {
  produces<ds50::Online::Positions> ();
}

ds50::OnlinePosition::~OnlinePosition() {}

void ds50::OnlinePosition::produce(art::Event & e) {
  art::Handle<ds50::Online::Clusters> clusters;
  e.getByLabel ("cluster", clusters); // check eventuale
  if (!clusters.isValid ()) return;


  std::unique_ptr<ds50::Online::Positions> positions(new ds50::Online::Positions);
  ds50::Online::Position current_position;

  for (auto const &cl: clusters->clusters) {
    current_position.z = drift_ * cl.start_time_ns;

    float total_charge = 0;
    for (int i = 0; i < 40; i++) 
      if (cds_[i].z > 0 && cl.pmts[i].abs_charge >= 0) {
	float c = cl.pmts[i].charge;
	if (c < 0.5) c = 0.; // ignore small or negative charges
	cds_[i].c = c;
	total_charge += c;
      } else cds_[i].c = 0;

    for (int i = 0; i < 40; i++) current_position.normalized_charge[i] = cds_[i].nc = cds_[i].c / total_charge;

    // center of light (barycenter)
    current_position.x_col = current_position.y_col = 0;
    for (int i = 0; i < 40; i++) {
      current_position.x_col += cds_[i].x * cds_[i].nc;
      current_position.y_col += cds_[i].y * cds_[i].nc;
    }

    // least square
    if (do_lkl_) {
      TF2 f("f", this, -4.5, 4.5, -4.5, 4.5, 0, "position_lkl");
      double x = current_position.x_col - 100, y = current_position.x_col, _lkl;
      try {
	f.GetMinimumXY(x, y);
	gMinuit->Command("SET NOW");
	_lkl = lkl (x, y);
      } catch (...) { x = y = _lkl = -20; }
      current_position.x_ls = x;
      current_position.y_ls = y;
      current_position.lkl_ls = _lkl;
    }

    positions->positions.push_back(current_position);
  }

  e.put(std::move(positions));
}

float ds50::OnlinePosition::lkl (float x, float y) const {
  if (x * x + y * y > TPC_R * TPC_R) return 10000;

  std::vector<float> sa(40, 0.);

  for (size_t i = 0; i < 40; i++) if (cds_[i].c > 0) {
    int id = (cds_[i].r > 0) ? int(cds_[i].r) - 1 : 0;

    float d2t = ((cds_[i].x - x) * (cds_[i].x - x) + (cds_[i].y - y) * (cds_[i].y - y)) / TPC_R / TPC_R;
    float p = avoid_tgraph_insane_extrapolation (d2t, *topology_corrections_[id]);
#ifdef SA_RINGS
    sa[i] = avoid_tgraph_insane_extrapolation (1 - p, *solid_angles_[id]);
#else
    sa[i] = avoid_tgraph_insane_extrapolation (1 - p, *solid_angles_[i]);
#endif
//      float d2r = distance_ * distance_ + d2t; // * (1 - std::exp (-d2t / R / R / 2));
//      sa[i] = (1 - sqrt(d2r) / sqrt (d2r + PMT_R * PMT_R * sqrt (distance_ / sqrt(d2r)))) / (1 - distance_ / sqrt (distance_ * distance_ + PMT_R * PMT_R));
  }

//  std::cout << "total solid angle " << std::accumulate(sa.begin(), sa.end(), 0.) << std::endl;

  float ll = 0;
  for (size_t i = 0; i < 40; i++) if (cds_[i].c > 0) ll += (sa[i] - cds_[i].nc) * (sa[i] - cds_[i].nc);
  return ll;
}

float ds50::OnlinePosition::avoid_tgraph_insane_extrapolation (float x, const TGraph& g) const {
  if (x <= g.GetX ()[0]) return g.GetY ()[0];
  int n = g.GetN ();
  if (x >= g.GetX ()[n - 1]) return g.GetY ()[n - 1];

  return g.Eval (x);
}

void ds50::OnlinePosition::beginRun(art::Run &) {
  art::ServiceHandle<ds50::DBInterface> dbi;
  ds50::db::result r = dbi->latest ("calib.pmt_geometry", "");
  if (r.cell_elements("r", 0) != 40) throw cet::exception("OnlinePosition") << "PMT geometry did not return 40 elements";
  for (int i = 0; i < 40; i++) {
   cds_[i].x = r.get<float>("x", 0, i);
   cds_[i].y = r.get<float>("y", 0, i);
   cds_[i].z = r.get<float>("z", 0, i);
   cds_[i].r = r.get<float>("r", 0, i);
  }

  float radii[] = { 0, 2 / TPC_R, 3.46 / TPC_R, 4 / TPC_R };
#ifdef SA_RINGS
  std::string labels[] = { "h0", "h2", "h3", "h4" };
#endif

  TFile f(sa_calib_.c_str ());
  if (f.IsZombie ()) throw cet::exception("OnlinePosition") << "can not open solid angle calibration file " << sa_calib_;
#ifdef SA_RINGS
  for (int i = 0; i < 4; i++) {
    TH1 *h = dynamic_cast<TH1*>(f.Get (labels[i].c_str ()));
#else
  for (int i = 0; i < 40; i++) {
    TH1 *h = dynamic_cast<TH1*>(f.Get (Form ("h%d", i)));
#endif
    solid_angles_[i] = inverse_cumulative (h);
  }
  f.Close ();

  for (int i = 0; i < 4; i++) {
    TH1F pc ("pc", "", 800, 0, 4);
    for (float x = -1; x < 1; x += 5e-4) {
      for (float y = -1; y < 1; y += 5e-4) {
	if (x * x + y * y > 1) continue;
	pc.Fill (x * x + (y - radii[i]) * (y - radii[i]));
      }
    }
    topology_corrections_[i] = cumulative (&pc);
  }
}

std::unique_ptr<TGraph> ds50::OnlinePosition::inverse_cumulative (TH1* h) {
  h->Scale (1. / h->Integral ());

  std::unique_ptr<float[]> x(new float[h->GetNbinsX ()]), y(new float[h->GetNbinsX ()]); 
  double sum = 0;
  int valid_entries = 0;
  for (; valid_entries < h->GetNbinsX (); valid_entries ++) {
    y[valid_entries] = sum;
    x[valid_entries] = h->GetBinLowEdge (valid_entries + 1);
    if (sum >= (1 - 1e-6)) break;
    sum += h->GetBinContent (valid_entries + 1);
  }
  return std::unique_ptr<TGraph> (new TGraph (valid_entries + 1, y.get (), x.get ()));
}

std::unique_ptr<TGraph> ds50::OnlinePosition::cumulative (TH1* h) {
  h->Scale (1. / h->Integral ());

  std::unique_ptr<float[]> x(new float[h->GetNbinsX ()]), y(new float[h->GetNbinsX ()]); 
  double sum = 0;
  int valid_entries = 0;
  for (; valid_entries < h->GetNbinsX (); valid_entries ++) {
    y[valid_entries] = sum;
    x[valid_entries] = h->GetBinLowEdge (valid_entries + 1);
    if (sum >= (1 - 1e-6)) break;
    sum += h->GetBinContent (valid_entries + 1);
  }
  return std::unique_ptr<TGraph> (new TGraph (valid_entries + 1, x.get (), y.get ()));
}


DEFINE_ART_MODULE(ds50::OnlinePosition)
